##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import os
import sys
import exputils as eu
from glob import glob
import re
import numpy as np
import warnings
import collections
import importlib.util

# TODO: Feature - allow to load data from several campaigns

def load_experiment_descriptions(experiments_directory=None,
                                 allowed_experiments_id_list=None,
                                 denied_experiments_id_list=None,
                                 experiment_directory_template=None,
                                 repetition_directory_template=None):

    if experiments_directory is None:
        experiments_directory = os.path.join('..', eu.DEFAULT_EXPERIMENTS_DIRECTORY)

    if allowed_experiments_id_list is not None and denied_experiments_id_list is not None:
        raise ValueError('allowed_experiments_id_list and denied_experiments_id_list can not be set at the same time!')

    if experiment_directory_template is None: experiment_directory_template = eu.EXPERIMENT_DIRECTORY_TEMPLATE
    experiment_directory_template = re.sub('\{.*\}', '*', experiment_directory_template)

    if repetition_directory_template is None: repetition_directory_template = eu.REPETITION_DIRECTORY_TEMPLATE
    repetition_directory_template = re.sub('\{.*\}', '*', repetition_directory_template)

    experiment_descriptions = eu.AttrDict()

    exp_directories = glob(os.path.join(experiments_directory, experiment_directory_template))
    for order, exp_directory in enumerate(np.sort(exp_directories)):

        try:
            exp_id = re.findall(r'\d+', os.path.basename(exp_directory))[0]
        except IndexError as err:
            raise ValueError('The experiments_directory (\'{}\') seems not to have experiment folders!'.format(experiments_directory)) from err

        is_add_experiment_descr = True
        if allowed_experiments_id_list is not None and exp_id not in allowed_experiments_id_list:
            is_add_experiment_descr = False
        elif denied_experiments_id_list is not None and exp_id in denied_experiments_id_list:
            is_add_experiment_descr = False

        if is_add_experiment_descr:
            experiment_descr = eu.AttrDict()
            experiment_descr.id = exp_id
            experiment_descr.name = 'exp {}'.format(exp_id)
            experiment_descr.order = order
            experiment_descr.is_load_data = True
            experiment_descr.directory = exp_directory
            experiment_descr.short_name = 'e{}'.format(exp_id)
            experiment_descr.description = ''

            # find repetition directories and ids
            repetition_directories = glob(os.path.join(exp_directory, repetition_directory_template))
            experiment_descr.repetition_directories = repetition_directories
            if experiment_descr.repetition_directories:
                experiment_descr.repetition_directories.sort()

            experiment_descr.repetition_ids = []
            for rep_directory in np.sort(repetition_directories):
                rep_id = re.findall(r'\d+', os.path.basename(rep_directory))[0]
                experiment_descr.repetition_ids.append(int(rep_id))
            experiment_descr.repetition_ids.sort()

            experiment_descriptions[exp_id] = experiment_descr

    return experiment_descriptions


def load_experiment_data(experiment_descriptions=None, experiments_directory=None, allowed_experiments_id_list=None, denied_experiments_id_list=None,
                         data_directory=None, is_load_repetition_data=True,
                         pre_allowed_data_filter=None, pre_denied_data_filter=None, post_allowed_data_filter=None, post_denied_data_filter=None,
                         on_experiment_data_loaded=None, on_repetition_data_loaded=None):
    """
    Loads experimental data from experiments and their repetitions.

    :param experiments_directory:  Path to the directory that contains the experiments. (default = './experiments')
    :param allowed_experiments_id_list: List of experiment ids that are only allowed to be loaded, not other will be loaded. (default = None)
    :param denied_experiments_id_list: List of experiment ids that will not be loaded, all other will. (default = None)
    :param experiment_descriptions: Descriptions about the experiments that should be loaded. If it contains a is_load_data property, then this
                                    is checked if the experiment should be loaded or not. (default = None)
    :param data_directory: Relative path of the data directories under the experiments and repetitions. (default = '/data')
    :param is_load_repetition_data: Should the data of repetitions be loaded. (default = True)
    :param pre_allowed_data_filter: List of data items that should only be loaded before the on_experiment_data_loaded and on_repetition_data_loaded
                                    functions are called. (default = None)
    :param pre_denied_data_filter: List of data items that should not be loaded before the on_experiment_data_loaded and on_repetition_data_loaded
                                   functions are called. (default = None)
    :param post_allowed_data_filter: List of data items that are only allowed after the on_experiment_data_loaded and on_repetition_data_loaded
                                     functions are called. (default = None)
    :param post_denied_data_filter: List of data items that are removed after the on_experiment_data_loaded and on_repetition_data_loaded
                                    functions are called. (default = None)
    :param on_experiment_data_loaded: List of functions that will be called after experiment data is loaded. (default = None)
                                      Form of the functions: func(experiment_id, experiment_data)
    :param on_repetition_data_loaded: List of functions that will be called after repetition data is loaded. (default = None)
                                      Form of the functions: func(experiment_id, repetition_id, repetition_data)
    :return: Data and experiment descriptions.

    """

    if experiments_directory is not None and experiment_descriptions is not None:
        raise ValueError('Can not set experiment_directory and experiment_descriptions at the same time!')

    if experiment_descriptions is not None and (allowed_experiments_id_list is not None or denied_experiments_id_list is not None):
        raise ValueError('experiment_descriptions and (allowed_experiments_id_list or denied_experiments_id_list) can not be set at the same time!')

    if allowed_experiments_id_list is not None and denied_experiments_id_list is not None:
        raise ValueError('allowed_experiments_id_list and denied_experiments_id_list can not be set at the same time!')

    if experiment_descriptions is None:
        experiment_descriptions = load_experiment_descriptions(
                experiments_directory=experiments_directory,
            allowed_experiments_id_list=allowed_experiments_id_list,
            denied_experiments_id_list=denied_experiments_id_list
        )
    else:
        experiment_descriptions = experiment_descriptions

    if on_experiment_data_loaded is None:
        on_experiment_data_loaded = []

    if on_repetition_data_loaded is None:
        on_repetition_data_loaded = []

    # load experiments according to the order in the experiment_descriptions
    sorted_experiment_ids = eu.data.get_ordered_experiment_ids_from_descriptions(experiment_descriptions)

    data = collections.OrderedDict()
    for exp_id in sorted_experiment_ids:
        exp_descr = experiment_descriptions[exp_id]

        if 'is_load_data' not in exp_descr or exp_descr['is_load_data']:
            try:
                data[exp_id] = load_single_experiment_data(
                    exp_descr['directory'],
                    data_directory=data_directory,
                    allowed_data_filter=pre_allowed_data_filter,
                    denied_data_filter=pre_denied_data_filter)

                for callback_function in on_experiment_data_loaded:
                    callback_function(exp_id, data[exp_id])

                _filter_data(data[exp_id], post_allowed_data_filter, post_denied_data_filter)

            except FileNotFoundError:
                if not exp_descr.repetition_ids or not is_load_repetition_data:
                    warnings.warn('Could find data for experiment {!r} ({!r}). Skipped ...'.format(exp_id, exp_descr['directory']))

            except Exception as e:
                raise Exception('Exception during loading of data for experiment {!r} ({!r})!'.format(exp_id, exp_descr['directory'])) from e

            # load data of each repetition
            if is_load_repetition_data:
                if eu.REPETITION_DATA_KEY in data:
                    warnings.warn('A statistic called {!r} was loaded for experiment data. Can not store repetition data under the same data source name. Skip to load repetition data. Please rename this statistic.'.format(eu.REPETITION_DATA_KEY))
                else:
                    cur_rep_statistics_dict = dict()
                    for rep_id in exp_descr.repetition_ids:
                        cur_rep_directory = os.path.join(exp_descr['directory'], eu.REPETITION_DIRECTORY_TEMPLATE.format(rep_id))
                        try:
                            cur_rep_statistics_dict[rep_id] = load_single_experiment_data(
                                cur_rep_directory,
                                data_directory=data_directory,
                                allowed_data_filter=pre_allowed_data_filter,
                                denied_data_filter=pre_denied_data_filter)

                            for callback_function in on_repetition_data_loaded:
                                callback_function(exp_id, rep_id, cur_rep_statistics_dict[rep_id])

                            _filter_data(cur_rep_statistics_dict[rep_id], post_allowed_data_filter, post_denied_data_filter)

                        except FileNotFoundError:
                            warnings.warn('Could not find data for repetition {} of experiment {!r} ({!r}). Skipped ...'.format(rep_id, exp_id, exp_descr['directory']))

                        except Exception as e:
                            raise Exception('Exception during loading of data for repetition {} of experiment {!r} ({!r})!'.format(rep_id, exp_id, exp_descr['directory'])) from e

                    if cur_rep_statistics_dict:
                        # in case no experimental level data exists
                        if exp_id not in data:
                            data[exp_id] = eu.AttrDict()

                        data[exp_id][eu.REPETITION_DATA_KEY] = cur_rep_statistics_dict

    return data, experiment_descriptions


def load_single_experiment_data(experiment_directory, data_directory=None, allowed_data_filter=None, denied_data_filter=None, allow_pickle=True):
    """Loads the data from a single experiment."""

    if data_directory is None:
        data_directory = eu.DEFAULT_DATA_DIRECTORY

    # need to allow also logging to be able to load data that is in logging.npz files
    if allowed_data_filter is not None:
        allowed_data_filter.append('logging')

    statistics = eu.io.load_numpy_files(
        os.path.join(experiment_directory, data_directory),
        allowed_data_filter=allowed_data_filter,
        denied_data_filter=denied_data_filter,
        allow_pickle=allow_pickle)

    # TODO: Refactor - make loading of npz files without the 'logging' sub-directory as a general cases
    if 'logging' in statistics:
        statistics.update(statistics['logging'])
        del statistics['logging']

    return statistics


def _filter_data(data, allowed_data_list, denied_data_list):
    # get the data_elements that should be deleted
    delete_keys = [k for k in data.keys() if not eu.misc.is_allowed(k, allowed_list=allowed_data_list, denied_list=denied_data_list)]
    for delete_key in delete_keys:
        del data[delete_key]


def load_experiment_python_module(module_path, experiment_id=None, repetition_id=None, experiments_directory=None, exec_module=True,
                                  experiment_directory_template=None, repetition_directory_template=None,
                                  add_execution_directory_to_sys_path=True):
    """
    Loads a python module that is under a certain repetition of an experiment. Can be used to load for example the configuration of a repetition.

    Example:
        module = load_experiment_python_module('repetition_config.py', experiment_id=100, repetition_id=3)

    :param module_path: Name of the python module (including the '.py' ending).
    :param experiment_id: ID of the experiment. If not provided, then the module is assumed to be under the experiments directory.
    :param repetition_id: ID of the repetition. If not provided, then the module is assumed to be under the experiment directory.
    :param experiments_directory: Path to the experiments directory. (Default = '../experiments')
    :param exec_module: True if the module should be executed. (Default = True)
    :param experiment_directory_template: Alternative template string for experiment directories. (Default = 'experiment_{}')
    :param repetition_directory_template: Alternative template string for repetition directories. (Default = 'repetition_{}')
    :param add_execution_directory_to_sys_path: Should the directory in which the code was executed (experiment or repetition) be temporaily added
                                                to the python system path. This might be necessary if the loaded object has references to it.
                                                (Default = True)
    :return: Loaded module.
    """

    if experiments_directory is None:
        experiments_directory = os.path.join('..', eu.DEFAULT_EXPERIMENTS_DIRECTORY)

    full_module_path = experiments_directory

    # only add experiment subfolder if needed
    if experiment_id is not None:

        if experiment_directory_template is None:
            experiment_directory_template = eu.EXPERIMENT_DIRECTORY_TEMPLATE

        experiment_directory = experiment_directory_template.format(experiment_id)

        full_module_path = os.path.join(full_module_path, experiment_directory)

        # only add repetition subfolder if needed
        if repetition_id is not None:

            if repetition_directory_template is None:
                repetition_directory_template = eu.REPETITION_DIRECTORY_TEMPLATE

            repetition_directory = repetition_directory_template.format(repetition_id)

            full_module_path = os.path.join(full_module_path, repetition_directory)

    # construct the full path to the module
    full_module_path = os.path.join(full_module_path, module_path)

    filename = os.path.basename(module_path)
    module_name = filename.replace('.py', '')

    spec = importlib.util.spec_from_file_location(module_name, full_module_path)

    # creates a new module based on spec
    module = importlib.util.module_from_spec(spec)

    if exec_module:
        # add the directory in which the code was executed to system path
        if add_execution_directory_to_sys_path:
            sys.path.append(os.path.dirname(full_module_path))

        spec.loader.exec_module(module)

        if add_execution_directory_to_sys_path:
            sys.path.pop()

    return module


def load_experiment_data_single_object(name, experiment_id=None, repetition_id=None, experiments_directory=None, data_directory=None,
                                       experiment_directory_template=None, repetition_directory_template=None,
                                       add_execution_directory_to_sys_path=True):
    """
    Loads a object (a dill file) that is saved in the data directory of an experiment or repetition. Such objects are saved using the
    log.add_single_object() function.geht es

    Example:
        log.add_single_object('my_object', obj)
        obj = load_experiment_data_single_object('my_object', experiment_id=100, repetition_id=3)

    :param name: Name of the object or the dill file (including the '.dill' ending is optional).
    :param experiment_id: ID of the experiment. If not provided, then the object is assumed to be under the experiments data directory.
    :param repetition_id: ID of the repetition. If not provided, then the object is assumed to be under the experiment data directory.
    :param experiments_directory: Path to the experiments directory. (Default = '../experiments')
    :param data_directory: Name of the data directory. (Default = 'data')
    :param experiment_directory_template: Alternative template string for experiment directories. (Default = 'experiment_{}')
    :param repetition_directory_template: Alternative template string for repetition directories. (Default = 'repetition_{}')
    :param add_execution_directory_to_sys_path: Should the directory in which the code was executed (experiment or repetition) be temporaily added
                                                to the python system path. This might be necessary if the loaded object has references to it.
                                                (Default = True)
    :return: Loaded objected.
    """


    if experiments_directory is None:
        experiments_directory = os.path.join('..', eu.DEFAULT_EXPERIMENTS_DIRECTORY)

    full_execution_dir_path = experiments_directory

    # only add experiment subfolder if needed
    if experiment_id is not None:

        if experiment_directory_template is None:
            experiment_directory_template = eu.EXPERIMENT_DIRECTORY_TEMPLATE

        experiment_directory = experiment_directory_template.format(experiment_id)

        full_execution_dir_path = os.path.join(full_execution_dir_path, experiment_directory)

        # only add repetition subfolder if needed
        if repetition_id is not None:

            if repetition_directory_template is None:
                repetition_directory_template = eu.REPETITION_DIRECTORY_TEMPLATE

            repetition_directory = repetition_directory_template.format(repetition_id)

            full_execution_dir_path = os.path.join(full_execution_dir_path, repetition_directory)

    if data_directory is None:
        data_directory = eu.DEFAULT_DATA_DIRECTORY

    # construct the full path to the module
    full_dill_path = os.path.join(full_execution_dir_path, data_directory, name)

    # add the directory in which the code was executed to system path
    if add_execution_directory_to_sys_path:
        sys.path.append(full_execution_dir_path)

    obj = eu.io.dill.load_dill(full_dill_path)

    if add_execution_directory_to_sys_path:
        sys.path.pop()

    return obj



##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
import numpy as np
from tabulate import tabulate as original_tabulate

def tabulate_meanstd(data=None, config=None, **kwargs):
    """
    Plots a table over several datasource.

    :param data:
    :param config:
        tabulate: Parameters for the tabulate function that plots the table.
                  See https://pypi.org/project/tabulate/ for possible parameters.
                  Some important ones:

            tablefmt: Format of the table. (Default='html')
                        "plain", "simple", "github", "grid", "fancy_grid", "pipe", "orgtbl", "jira", "presto", "pretty",
                        "psql",  "rst", "mediawiki", "moinmoin", "youtrack", "html", "unsafehtml", "latex", "latex_raw",
                        "latex_booktabs", "latex_longtable", "textile", "tsv"

    :param kwargs:
    :return:
    """

    default_config = eu.AttrDict(

        primary_content_function = np.nanmean,

        secondary_content_function = np.nanstd,

        flip_rows_and_cols = False,

        tabulate=eu.AttrDict(
            tablefmt='html', #
            numalign='right',
        ),

        cell_format = '{:.3f} ({:.3f})',

        top_left_cell_content = '',

        default_row_label = '<row_idx>',
        default_column_label = '<column_idx>',

        labels=[],  # holds all labels in a specific structure

    )
    config = eu.combine_dicts(kwargs, config, default_config)

    # remove the secondary information from the cell_format if the secondary information is set to None and the cell_format was not changed
    if config.secondary_content_function is None and config.cell_format == default_config.cell_format:
        config.cell_format = '{}'

    if data is None:
        data = np.array([])

    # format data in form [rows][columns][elems_per_trace:numpy.ndarray]
    # subplot is a single table
    if isinstance(data, np.ndarray):
        data = [[data]]
    elif isinstance(data, list) and isinstance(data[0], np.ndarray):
        data = [data]
    elif not isinstance(data, list) and not isinstance(data[0], list) and not isinstance(data[0][0], np.ndarray):
        raise ValueError('Unsupported type of data!')

    # handle different input formats of labels
    if config.labels:
        # if only labels for columns are given, then add an empty label for the sub figure
        if isinstance(config.labels, list) and not isinstance(config.labels[0], tuple):
            config.labels = [('', config.labels)]

    row_headers = []
    column_headers = []

    primary_data = []
    secondary_data = []

    # interate over rows (subplots)
    for row_idx, row_data in enumerate(data):

        primary_data.append([])
        secondary_data.append([])

        # define label of the row
        if config.labels:
            row_label = config.labels[row_idx][0]
        else:
            row_label = config.default_row_label
        row_label = eu.misc.replace_str_from_dict(str(row_label), {'<row_idx>': row_idx})
        row_headers.append(row_label)

        # collect the data and labels for each trace
        for column_idx, cur_data in enumerate(row_data):

            # get column_header from labels of first column
            if row_idx == 0:

                # define label of the row
                if config.labels:
                    column_label = config.labels[row_idx][1][column_idx]
                    if isinstance(column_label, tuple):
                        column_label = column_label[0]
                else:
                    column_label = config.default_column_label
                column_label = eu.misc.replace_str_from_dict(str(column_label), {'<column_idx>': column_idx})
                column_headers.append(column_label)

            data_points = np.array([])

            if np.ndim(cur_data) == 0:
                data_points = np.array([cur_data])
            elif np.ndim(cur_data) == 1:
                data_points = cur_data
            else:
                # collect data over elements
                for elem_idx, elem_data in enumerate(cur_data):  # data elements

                    # get element data which could be in matrix format or array format
                    if np.ndim(elem_data) == 0:
                        cur_elem_data = np.array([elem_data])
                    elif np.ndim(elem_data) == 1:
                        cur_elem_data = elem_data
                    elif np.ndim(elem_data) == 2:
                        if elem_data.shape[0] == 1:
                            cur_elem_data = elem_data[0, :]
                        elif elem_data.shape[1] == 1:
                            cur_elem_data = elem_data[1, 0]
                        else:
                            raise ValueError('Invalid data format!')
                    else:
                        raise ValueError('Invalid data format!')

                    data_points = np.hstack((data_points, cur_elem_data))

            # try:
            if data_points[0] is not None:
                primary_data[row_idx].append(config.primary_content_function(data_points))

                if config.secondary_content_function:
                    secondary_data[row_idx].append(config.secondary_content_function(data_points))
                else:
                    secondary_data[row_idx].append(None)
            else:
                primary_data[row_idx].append(None)
                secondary_data[row_idx].append(None)



    # plot the results
    n_rows = len(primary_data)
    n_columns = len(primary_data[0])
    if config.flip_rows_and_cols:
        n_rows = len(primary_data[0])
        n_columns = len(primary_data)
        tmp = row_headers
        row_headers = column_headers
        column_headers = tmp

    table_content = [[None] * (n_columns + 1) for _ in range(n_rows + 1)]
    table_content[0][0] = config.top_left_cell_content


    # set row and column headers
    for row_idx in range(n_rows):
        table_content[row_idx + 1][0] = row_headers[row_idx]
    for column_idx in range(n_columns):
        table_content[0][column_idx + 1] = column_headers[column_idx]

    # fill table
    for row_idx in range(n_rows):
        for column_idx in range(n_columns):

            data_1_idx = row_idx
            data_2_idx = column_idx
            if config.flip_rows_and_cols:
                data_1_idx = column_idx
                data_2_idx = row_idx

            if primary_data[data_1_idx][data_2_idx] is None:
                cell_data = ''
            else:
                if isinstance(config.cell_format, str):
                    cell_data = config.cell_format.format(
                        primary_data[data_1_idx][data_2_idx],
                        secondary_data[data_1_idx][data_2_idx])
                else:
                    cell_data = eu.misc.call_function_from_config(
                        config.cell_format,
                        primary_data[data_1_idx][data_2_idx],
                        secondary_data[data_1_idx][data_2_idx])

            table_content[row_idx + 1][column_idx + 1] = cell_data

    table = original_tabulate(
        table_content,
        headers='firstrow',
        **config.tabulate)

    return table



##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
from exputils.manage.experimentgenerator import generate_experiment_files
from exputils.manage.experimentstarter import start_experiments
from exputils.manage.experimentstarter import start_slurm_experiments
from exputils.manage.experimentstarter import start_torque_experiments
from exputils.manage.experimentstarter import get_scripts
from exputils.manage.experimentstarter import get_number_of_scripts_to_execute
from exputils.manage.experimentstarter import get_number_of_scripts






##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
import numpy as np
import os
from glob import glob


def save_dict_to_numpy_files(data, path='.', mode = 'npy'):
    """
    Saves the data in a dictionary to numpy files. Either several npy or one npz (compressed or un-compressed).

    :param data: Dictionary with data.
    :param path: Path to folder if npy files are saved, or to the npz file.
    :param mode: Defines if npy files or one npz file are saved: 'npy', 'npz', 'cnpz' - compressed. (Default: 'npy')
    """

    # save logs in numpy format if they exist
    if mode.lower() == 'npy':
        eu.io.makedirs(path)
        for name, values in data.items():
            np.save(os.path.join(path, name), values)

    elif mode.lower() == 'npz':
        eu.io.makedirs_for_file(path)
        np.savez(path, **data)

    elif mode.lower() == 'cnpz':
        eu.io.makedirs_for_file(path)
        np.savez_compressed(path, **data)

    else:
        raise ValueError('Unknown numpy logging mode {!r}!'.format(mode))


def load_numpy_files(directory, allowed_data_filter=None, denied_data_filter=None, allow_pickle=True):
    """Loads data from all npy and npz files in a given directory."""

    if allowed_data_filter is not None and denied_data_filter is not None:
        raise ValueError('in_data_filter and out_data_filter can not both be set, only one or none!')

    if not os.path.isdir(directory):
        raise FileNotFoundError('Directory {!r} does not exist!'.format(directory))

    data = eu.AttrDict()

    for file in glob(os.path.join(directory, '*.npy')):
        stat_name = os.path.splitext(os.path.basename(file))[0]

        if eu.misc.is_allowed(stat_name, allowed_list=allowed_data_filter, denied_list=denied_data_filter):
            try:
                stat_val = np.load(file, allow_pickle=allow_pickle)
            except FileNotFoundError:
                raise
            except Exception as e:
                raise Exception('Exception during loading of file {!r}!'.format(file)) from e

            if len(stat_val.shape) == 0:
                stat_val = stat_val.dtype.type(stat_val)

            data[stat_name] = stat_val

    for file in glob(os.path.join(directory, '*.npz')):
        stat_name = os.path.splitext(os.path.basename(file))[0]
        if eu.misc.is_allowed(stat_name, allowed_list=allowed_data_filter, denied_list=denied_data_filter):
            try:
                stat_vals = eu.AttrDict(np.load(file, allow_pickle=allow_pickle))
            except FileNotFoundError:
                raise
            except Exception as e:
                raise Exception('Exception during loading of file {!r}!'.format(file)) from e

            # remove data that should not be loaded
            keys = [k for k, v in stat_vals.items() if not eu.misc.is_allowed(k, allowed_list=allowed_data_filter, denied_list=denied_data_filter)]
            for x in keys:
                del stat_vals[x]

            # numpy encapsulates scalars as darrays with an empty shape
            # recover the original type
            for substat_name, substat_val in stat_vals.items():
                if len(substat_val.shape) == 0:
                    stat_vals[substat_name] = substat_val.dtype.type(substat_val)

            data[stat_name] = stat_vals

    return data





##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import ipywidgets
import exputils as eu

# TODO: when creating a new code cell, also save directly the selection of experiments and repetitions for it

class CodeProducerWidget(ipywidgets.Box):

    @staticmethod
    def default_config():
        dc = eu.AttrDict()

        dc.code_templates = []

        dc.main_box = eu.AttrDict()

        dc.title = 'Code Production'
        dc.accordion = eu.AttrDict(
            selected_index=None,
            layout=eu.AttrDict(width='99%'))
        dc.accordion_box = eu.AttrDict(
            layout=eu.AttrDict(
                width='99%',
                display='flex',
                flex_flow='column',
                align_items='stretch'))
        dc.default_button = eu.AttrDict(
            layout=eu.AttrDict(
                width='95%'))

        return dc

    def __init__(self, code_variables, config=None, **kwargs):
        self.config = eu.combine_dicts(kwargs, config, self.default_config())
        super().__init__(cildren=[], **self.config.main_box)

        self.code_variables = code_variables

        if isinstance(self.config.code_templates, dict): self.config.code_templates = [self.config.code_templates]

        child_widgets = []
        if len(self.config.code_templates) == 1:
            # only create a button
            btn = self._create_button(
                1,
                self.config.code_templates[0],
                default_label=self.config.title)
            child_widgets.append(btn)
        else:

            accordion_child_widgets = []
            for btn_idx, btn_descr in enumerate(self.config.code_templates):
                btn = self._create_button(
                    btn_idx + 1,
                    btn_descr)
                accordion_child_widgets.append(btn)

            accordion_box = ipywidgets.Box(
                children=accordion_child_widgets,
                **self.config.accordion_box)

            accordion = ipywidgets.Accordion(
                children=[accordion_box],
                **self.config.accordion)
            accordion.set_title(0, self.config.title)

            child_widgets.append(accordion)

        eu.gui.jupyter.add_children_to_widget(self, child_widgets)


    def _create_button(self, idx, btn_descr, default_label=None):

        if default_label is None:
            default_label = 'Code {}'.format(idx)

        if isinstance(btn_descr, str):
            # just the template as string
            btn_label = default_label
            btn_tooltip = ''
            btn_code_template = btn_descr
        else:
            # dictionary with description
            btn_label = btn_descr.get('name', default_label)
            btn_tooltip = btn_descr.get('tooltip', '')
            btn_code_template = btn_descr['code_template']

        btn = ipywidgets.Button(
            description=btn_label,
            tooltip=btn_tooltip,
            **self.config.default_button)
        btn.code_template = btn_code_template

        btn.on_click(self.on_click_handler)

        return btn


    def on_click_handler(self, btn):

        code_template = btn.code_template

        if isinstance(self.code_variables, dict):
            code_variables = self.code_variables
        else:
            code_variables = self.code_variables()

        code = eu.misc.replace_str_from_dict(
            code_template,
            code_variables,
            pattern_format='<{}>')

        eu.gui.jupyter.create_new_cell(code)


##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
import os
import IPython
from datetime import datetime
import ipywidgets
import base64
from IPython.display import Javascript, display

def create_new_cell(code='', where='below'):
    """Create a code cell in the IPython Notebook.

    Parameters
    code: unicode
        Code to fill the new code cell with.
    where: unicode
        Where to add the new code cell.
        Possible values include:
            at_bottom
            above
            below"""
    encoded_code = (base64.b64encode(str.encode(code))).decode()

    out = ipywidgets.Output()
    with out:
        IPython.display.display_javascript("""
                var code = IPython.notebook.insert_cell_{0}('code');
                code.set_text(atob("{1}"));
            """.format(where, encoded_code), raw=True)
    out.clear_output()


def save_config(config, name, directory=None, profile=None):
    '''
    Saves a given config object as a JSON. The config is usually a dictionary.

    :param name: Name of the configuration.
    :param config: Dictionary with the configuration.
    :param directory:  Optional directory where config is stored. (Default: './.ipython_config/').
    :param profile: Optional profile under which configuration is stored. (Default: name of the current notebook file).
    :return:
    '''

    if directory is None: directory = eu.gui.jupyter.DEFAULT_CONFIG_DIRECTORY
    if profile is None: profile = eu.gui.jupyter.get_notebook_name()

    filepath = os.path.join(directory, profile, name + '.json')

    eu.io.save_dict_as_json_file(config, filepath)


def load_config(name, directory=None, profile=None):
    '''
    Loads a configuration that is stored as JSON file as a AttrDict (dictionary).
    Return an empty AttrDict (dictionary) if the configuration does not exists.

    :param name: Name of the configuration.
    :param directory:  Optional directory where config is stored. (Default: './.ipython_config/').
    :param profile: Optional profile under which configuration is stored. (Default: name of the current notebook file).
    :return: AttrDict (dictionary) with config.
    '''

    if directory is None: directory = eu.gui.jupyter.DEFAULT_CONFIG_DIRECTORY
    if profile is None: profile = eu.gui.jupyter.get_notebook_name()

    filepath = os.path.join(directory, profile, name + '.json')

    try:
        loaded_dict = eu.io.load_dict_from_json_file(filepath)
    except FileNotFoundError:
        loaded_dict = {}

    return eu.AttrDict(loaded_dict)


def remove_children_from_widget(widget, idxs):
    '''
    Removes children widgets from a given widget according to their index.

    :param widget: Widget from whcih the children should be removed.
    :param idxs: Indexes (or single index) of child widgets that should be removed.
    '''

    if not isinstance(idxs, list):
        idxs = [idxs]

    children = list(widget.children)

    idxs = idxs.copy()
    idxs.sort(reverse=True)
    for idx in idxs:
        del children[idx]

    widget.children = tuple(children)


def add_children_to_widget(widget, children, idx=None):
    '''
    Adds children widgets to the given widget.

    :param widget: ipywidget to whoms children attribute the given children should be added.
    :param children: Child widgets or list of child widgets that should be added to the given widget.
    :param idx: Position at which the given children should be added. If None is given then at the end. (default = None)
    '''

    if not isinstance(children, list):
        children = [children]

    existing_children_list = list(widget.children)

    if idx is None:
        new_children_list = existing_children_list + children
    else:
        new_children_list = existing_children_list
        new_children_list[idx:idx] = children

    widget.children = tuple(new_children_list)


def set_children_of_widget(widget, idxs, children):
    children_widgets =  list(widget.children)
    children_widgets[idxs] = children
    widget.children = tuple(children_widgets)


def generate_random_state_backup_name():
    time_in_sec = int(datetime.now().timestamp() - datetime(2020, 1, 1).timestamp())
    return 'state_backup_{}'.format(time_in_sec)




##
## This file is part of the exputils package.
##
## Copyright: INRIA
## Year: 2022
## Contact: chris.reinke@inria.fr
##
## exputils is provided under GPL-3.0-or-later
##
import exputils as eu
import ipywidgets


class MultiSelectionWidget(ipywidgets.VBox):

    @staticmethod
    def default_config():
        dc = eu.AttrDict()

        dc.title = 'SELECTED_CHOICES'  # SELECTED_CHOICES, TITLE_LABELS

        # labels for each possible choice that are then shown in the title if the 'TITLE_LABELS'
        # option is used for the title - configuration
        dc.title_labels = []

        dc.is_select_all_choice = True
        dc.all_selected_title_label = 'all'

        # defines if an extra choice called none is given
        dc.is_select_none_choice = False

        dc.none_selected_title_label = 'none'

        dc.main_vbox = eu.AttrDict()

        dc.checkboxes = eu.AttrDict()

        dc.accordion = eu.AttrDict()

        dc.multi_checkbox_vbox = eu.AttrDict(layout=eu.AttrDict(overflow='scroll'))

        return dc

    def __init__(self, choices=None, config=None, **kwargs):
        self.config = eu.combine_dicts(kwargs, config, self.default_config())

        if choices is None: choices = []

        self.choices = choices

        self._on_selection_changed_event_handlers = []

        # define if the events for checkboxes are active
        self._is_on_select_all_checkbox_changed_active = True
        self._is_on_select_none_checkbox_changed_active = True
        self._is_on_checkbox_changed_active = True

        self._is_selection_changed_event_active = True

        # create checkboxes
        self.choice_checkboxes = []
        for choice_idx, choice in enumerate(self.choices):
            # is_on = not self.config.is_select_all_choice

            chkbox = ipywidgets.Checkbox(
                description=str(choice),
                value=True,
                **self.config.checkboxes)

            chkbox.observe(self._on_checkbox_changed)

            self.choice_checkboxes.append(chkbox)


        checkboxes_list = self.choice_checkboxes

        if self.config.is_select_all_choice:
            self.select_all_checkbox = ipywidgets.Checkbox(
                description='all',
                value=True,
                **self.config.checkboxes)
            self.select_all_checkbox.observe(self._on_select_all_checkbox_changed)
            checkboxes_list = [self.select_all_checkbox] + checkboxes_list

        if self.config.is_select_none_choice:
            self.select_none_checkbox = ipywidgets.Checkbox(
                description='none',
                value=False,
                **self.config.checkboxes)
            self.select_none_checkbox.observe(self._on_select_none_checkbox_changed)
            checkboxes_list = [self.select_none_checkbox] + checkboxes_list

        self.multi_checkbox_vbox_widget = ipywidgets.VBox(
            children=checkboxes_list,
            **self.config.multi_checkbox_vbox)

        self.accordion_widget = ipywidgets.Accordion(
            children=[self.multi_checkbox_vbox_widget],
            selected_index=None,
            **self.config.accordion)
        self.accordion_widget.set_title(0, self.config.title)

        self._update_titel()

        super().__init__(children=[self.accordion_widget], **self.config.main_vbox)


    @property
    def is_all_selected(self):
        return len(self.selected_choices) == len(self.choices)


    @property
    def selected_choices(self):
        choices = []
        for chkbox_idx, chkbox in enumerate(self.choice_checkboxes):
            if chkbox.value:
                choices.append(self.choices[chkbox_idx])
        return choices


    @selected_choices.setter
    def selected_choices(self, selected_choices):

        old_is_selection_changed_event_active = self._is_selection_changed_event_active
        self._is_selection_changed_event_active = False

        if selected_choices == 'all':
            for chkbox in self.choice_checkboxes:
                chkbox.value = True
        elif selected_choices == 'none':
            for chkbox in self.choice_checkboxes:
                chkbox.value = False
        else:
            for choice_idx, choice in enumerate(self.choices):
                if choice in selected_choices:
                    self.choice_checkboxes[choice_idx].value = True
                else:
                    self.choice_checkboxes[choice_idx].value = False

        self._is_selection_changed_event_active = old_is_selection_changed_event_active
        self._call_selection_changed_event()


    @property
    def selected_choices_idxs(self):
        idxs = []
        for chkbox_idx, chkbox in enumerate(self.choice_checkboxes):
            if chkbox.value:
                idxs.append(chkbox_idx)
        return idxs


    @selected_choices_idxs.setter
    def selected_choices_idxs(self, selected_choices_idxs):

        old_is_selection_changed_event_active = self._is_selection_changed_event_active
        self._is_selection_changed_event_active = False

        for choice_idx in range(len(self.choices)):
            if choice_idx in selected_choices_idxs:
                self.choice_checkboxes[choice_idx].value = True
            else:
                self.choice_checkboxes[choice_idx].value = False

        self._is_selection_changed_event_active = old_is_selection_changed_event_active
        self._call_selection_changed_event()


    @property
    def selected_choices_inds(self):
        return [chkbox.value for chkbox in self.choice_checkboxes]


    @selected_choices_inds.setter
    def selected_choices_inds(self, selected_choices_inds):

        old_is_selection_changed_event_active = self._is_selection_changed_event_active
        self._is_selection_changed_event_active = False

        for choice_idx, is_selected in enumerate(selected_choices_inds):
            self.choice_checkboxes[choice_idx].value = is_selected

        self._is_selection_changed_event_active = old_is_selection_changed_event_active
        self._call_selection_changed_event()


    def on_selection_changed(self, handler):
        '''
        Register an event handler for changes of the selected choices.
        The handler receives a dict with information about the event.
        '''
        self._on_selection_changed_event_handlers.append(handler)


    def _update_titel(self):

        if self.config.title in ['SELECTED_CHOICES', 'TITLE_LABELS']:
            selected_choices_idxs = self.selected_choices_idxs

            if self.config.is_select_all_choice and self.select_all_checkbox.value == True:
                title = self.config.all_selected_title_label
            elif not selected_choices_idxs:
                title = self.config.none_selected_title_label
            else:
                if self.config.title == 'SELECTED_CHOICES':
                    labels = self.choices
                elif self.config.title == 'TITLE_LABELS':
                    labels = self.config.title_labels

                title = ''
                for idx, choice_idx in enumerate(selected_choices_idxs):
                    if idx > 0:
                        title += ', '
                    title += str(labels[choice_idx])

            self.accordion_widget.set_title(0, title)

    def _call_selection_changed_event(self):

        if self._is_selection_changed_event_active:

            self._update_titel()

            descr = {'name': 'selection_changed',
                     'new': {'value': False},
                     'owner': self,
                     'type': 'change'}

            for handler in self._on_selection_changed_event_handlers:
                handler(descr)


    def _on_select_all_checkbox_changed(self, event_descr):

        if self._is_on_select_all_checkbox_changed_active:

            # only react to changes of the value of a check box
            if event_descr['name'] == 'value' and event_descr['type'] == 'change':

                old_state = self._is_on_checkbox_changed_active
                self._is_on_checkbox_changed_active = False

                old_is_on_select_none_checkbox_changed_active = self._is_on_select_none_checkbox_changed_active
                self._is_on_select_none_checkbox_changed_active = False

                for chkbox in self.choice_checkboxes:
                    chkbox.value = event_descr['new']

                if self.config.is_select_none_choice:
                    self.select_none_checkbox.value = not event_descr['new']

                self._is_on_select_none_checkbox_changed_active = old_is_on_select_none_checkbox_changed_active
                self._is_on_checkbox_changed_active = old_state

                # call event, that the selection changed
                self._call_selection_changed_event()


    def _on_select_none_checkbox_changed(self, event_descr):

        if self._is_on_select_none_checkbox_changed_active:

            # only react to changes of the value of a check box
            if event_descr['name'] == 'value' and event_descr['type'] == 'change':

                old_state = self._is_on_checkbox_changed_active
                self._is_on_checkbox_changed_active = False

                old_is_on_select_all_checkbox_changed_active = self._is_on_select_all_checkbox_changed_active
                self._is_on_select_all_checkbox_changed_active = False

                for chkbox in self.choice_checkboxes:
                    chkbox.value = not event_descr['new']

                if self.config.is_select_all_choice:
                    self.select_all_checkbox.value = not event_descr['new']

                self._is_on_select_all_checkbox_changed_active = old_is_on_select_all_checkbox_changed_active
                self._is_on_checkbox_changed_active = old_state

                # call event, that the selection changed
                self._call_selection_changed_event()


    def _on_checkbox_changed(self, event_descr):

        if self._is_on_checkbox_changed_active:

            # only react to changes of the value of a check box
            if event_descr['name'] == 'value' and event_descr['type'] == 'change':

                # deactivate the event for "select all" checkbox
                # to not call this event again
                old_is_on_select_all_checkbox_changed_active = self._is_on_select_all_checkbox_changed_active
                self._is_on_select_all_checkbox_changed_active = False

                old_is_on_select_none_checkbox_changed_active = self._is_on_select_none_checkbox_changed_active
                self._is_on_select_none_checkbox_changed_active = False

                # all checkbox
                if event_descr['new'] == False and self.config.is_select_all_choice:
                    self.select_all_checkbox.value = False

                elif event_descr['new'] == True and self.config.is_select_all_choice:

                    # check if all choices are selected, if yes, then also activate the "select all" option
                    is_all_selected = True
                    for chkbox in self.choice_checkboxes:
                        if chkbox.value == False:
                            is_all_selected = False
                            break

                    if is_all_selected:
                        self.select_all_checkbox.value = True
                        # if self.config.is_select_none_choice:
                        #     self.select_none_checkbox.value = False

                # none checkbox
                if event_descr['new'] == True and self.config.is_select_none_choice:
                    self.select_none_checkbox.value = False

                elif event_descr['new'] == False and self.config.is_select_none_choice:

                    # check if no choices are selected, if yes, then also activate the "select none" option
                    is_none_selected = True
                    for chkbox in self.choice_checkboxes:
                        if chkbox.value == True:
                            is_none_selected = False
                            break

                    if is_none_selected:
                        self.select_none_checkbox.value = True
                        # if self.config.is_select_all_choice:
                        #     self.select_all_checkbox.value = False

                self._is_on_select_all_checkbox_changed_active = old_is_on_select_all_checkbox_changed_active
                self._is_on_select_none_checkbox_changed_active = old_is_on_select_none_checkbox_changed_active

                # call event, that the selection changed
                self._call_selection_changed_event()
